import unittest
import pathlib
import shutil
import os
import tempfile
import re
from pathlib import Path
from ontoenv import OntoEnv


class TestOntoEnvInit(unittest.TestCase):
    def tearDown(self):
        # Clean up any accidental .ontoenv in cwd
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def test_init_recreate_new_dir(self):
        with tempfile.TemporaryDirectory() as td:
            root = Path(td)
            env_path = root / "new_env"
            self.assertFalse(env_path.exists())
            env = OntoEnv(path=env_path, recreate=True)
            self.assertTrue((env_path / ".ontoenv").is_dir())
            sp = env.store_path()
            self.assertIsNotNone(sp)
            self.assertTrue(Path(sp).is_dir())

    def test_init_recreate_existing_empty_dir(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "empty_env"
            env_path.mkdir()
            self.assertTrue(env_path.is_dir())
            env = OntoEnv(path=env_path, recreate=True)
            self.assertTrue((env_path / ".ontoenv").is_dir())
            self.assertIsNotNone(env.store_path())

    def test_init_load_from_existing_dir(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "existing_env"
            env_path.mkdir()
            env = OntoEnv(path=env_path, recreate=True)
            env.flush()
            del env
            # load existing
            env2 = OntoEnv(path=env_path, read_only=False)
            self.assertEqual(env2.store_path(), str(env_path / ".ontoenv"))

    def test_init_recreate_existing_dir(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "existing_env"
            env_path.mkdir()
            env = OntoEnv(path=env_path, recreate=True)
            (env_path / ".ontoenv" / "dummy.txt").touch()
            self.assertTrue((env_path / ".ontoenv" / "dummy.txt").exists())
            # Recreate
            env = OntoEnv(path=env_path, recreate=True)
            self.assertTrue((env_path / ".ontoenv").is_dir())
            self.assertFalse((env_path / ".ontoenv" / "dummy.txt").exists())
            self.assertEqual(len(env.get_ontology_names()), 0)

    def test_init_read_only(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "existing_env"
            env_path.mkdir()
            env1 = OntoEnv(path=env_path, recreate=True)
            env1.close()
            env = OntoEnv(path=env_path, read_only=True)
            self.assertTrue((env_path / ".ontoenv").is_dir())
            with self.assertRaisesRegex(ValueError, "Cannot add to read-only store"):
                env.add("file:///dummy.ttl")

    def test_init_no_config_no_path_error(self):
        # Clean up potential leftover .ontoenv in cwd just in case
        if os.path.exists(".ontoenv"):
            if os.path.isfile(".ontoenv"):
                os.remove(".ontoenv")
            else:
                shutil.rmtree(".ontoenv")
        with self.assertRaisesRegex(ValueError, "OntoEnv directory not found at: \"./.ontoenv\""):
            OntoEnv()  # No args

    def test_init_path_no_env_error(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "no_env_here"
            env_path.mkdir()
            self.assertFalse((env_path / ".ontoenv").exists())
            # Be tolerant of macOS /private prefix differences by matching only the tail.
            tail_pattern = rf'OntoEnv directory not found at: "(.*/)?{re.escape(env_path.name)}/\.ontoenv"'
            with self.assertRaisesRegex(ValueError, tail_pattern):
                OntoEnv(path=env_path)

    def test_init_temporary(self):
        with tempfile.TemporaryDirectory() as td:
            env_path = Path(td) / "temp_env_root"
            env = OntoEnv(temporary=True, root=str(env_path), strict=False)
            self.assertFalse((env_path / ".ontoenv").exists())
            self.assertIsNone(env.store_path())
            try:
                env.add("http://example.com/nonexistent.ttl")
            except ValueError as e:
                self.assertNotIn("Cannot add to read-only store", str(e))
            except Exception:
                pass


if __name__ == "__main__":
    unittest.main()



from ontoenv import OntoEnv, version
from rdflib import Graph
print(version)


print("Make env")
env = OntoEnv(search_directories=["../brick"], strict=False, offline=False)
print(env)
print("get brick")
g = Graph()
env.get_closure("https://brickschema.org/schema/1.4-rc1/Brick", g)
print(len(g))

print("get brick 2")
brick = Graph()
brick.parse("../brick/Brick.ttl", format="turtle")
env.import_dependencies(brick)
print(len(brick))
brick_name = env.add("https://brickschema.org/schema/1.4/Brick.ttl")
print(f"Added {brick_name}")
del env

print("new env")
env2 = OntoEnv()
print(env2.store_path())

print("get brick again from URL")
brick = env2.get("https://brickschema.org/schema/1.4/Brick")
print(len(brick))
print(brick)
print(type(brick))

print("brick closure", env2.list_closure("https://brickschema.org/schema/1.4/Brick"))

env2.import_graph(brick, "https://w3id.org/rec")
brick.serialize("test.ttl", format="turtle")

print("qudtqk deps", env2.get_dependents('http://qudt.org/2.1/vocab/quantitykind'))

# get an rdflib.Dataset (https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html#rdflib.Dataset)
ds = env2.to_rdflib_dataset()
for graphname in ds.graphs():
    graph = ds.graph(graphname)
    print(f"Graph {graphname} has {len(graph)} triples")



#!/usr/bin/env python3
"""
Benchmark .ttl files vs rdf5d.

For each .ttl file in a directory, measures:
  - Size: original TTL vs generated .r5tu
  - Time: rdflib load/serialize vs r5tu build/stat (load)

Requires:
  - Python: rdflib (pip install rdflib)
  - r5tu CLI built with --features oxigraph (built automatically if not found)

Usage:
  python scripts/benchmark_ttl_dir.py DIR [-o results.csv] [--recursive]
                                         [--keep-artifacts]
                                         [--r5tu PATH] [--enable-mmap]

Notes:
  - The script tries to use target/release/r5tu if present; otherwise it will
    attempt to build it via `cargo build --release --features oxigraph[,mmap]`.
  - Load time for rdf5d is measured by timing `r5tu stat --file <file>`.
  - rdflib save time is for Graph.serialize() to Turtle.
"""

from __future__ import annotations

import argparse
import csv
import gc
import os
import shlex
import subprocess
import sys
import tempfile
import time
from pathlib import Path
import statistics as stats
from typing import Iterable, List, Optional, Tuple


def _r5tu_has_oxigraph_cli(r5tu_path: Path) -> bool:
    """Detect if the r5tu binary is built with the oxigraph CLI.

    The non-oxigraph build prints a message like:
      "r5tu CLI requires the 'oxigraph' feature. Try: ..."
    and does not expose clap subcommands.
    """
    try:
        proc = subprocess.run(
            [r5tu_path.as_posix(), "--help"], capture_output=True, text=True
        )
    except Exception:
        return False
    out = (proc.stdout or "") + (proc.stderr or "")
    if "requires the 'oxigraph' feature" in out:
        return False
    # Heuristic: help should mention subcommands like build-graph/stat
    return ("build-graph" in out) and ("stat" in out)


def find_or_build_r5tu(explicit_path: Optional[Path], enable_mmap: bool) -> Path:
    """Find existing r5tu binary or build it.

    Returns path to the binary, raising RuntimeError on failure.
    """
    if explicit_path:
        p = explicit_path.resolve()
        if not p.exists():
            raise RuntimeError(f"r5tu not found at {p}")
        if not _r5tu_has_oxigraph_cli(p):
            raise RuntimeError(
                f"r5tu at {p} does not expose the oxigraph-based CLI (build with --features oxigraph)"
            )
        if not _r5tu_supports_zstd(p):
            raise RuntimeError(
                f"r5tu at {p} does not support --zstd (build with --features zstd)."
            )
        return p

    # Prefer release build if present
    candidate = Path("target/release/r5tu")
    if candidate.exists() and _r5tu_has_oxigraph_cli(candidate) and _r5tu_supports_zstd(candidate):
        return candidate.resolve()

    # Try debug build
    candidate = Path("target/debug/r5tu")
    if candidate.exists() and _r5tu_has_oxigraph_cli(candidate) and _r5tu_supports_zstd(candidate):
        return candidate.resolve()

    # Build release with required features
    features = ["oxigraph", "zstd"]
    if enable_mmap:
        features.append("mmap")
    cmd = [
        "cargo",
        "build",
        "--release",
        "--features",
        ",".join(features),
    ]
    print("Building r5tu:", " ".join(shlex.quote(x) for x in cmd))
    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            "Failed to build r5tu. Ensure Rust toolchain is installed and network access is available for crates."
        ) from e

    candidate = Path("target/release/r5tu")
    if not candidate.exists() or not _r5tu_has_oxigraph_cli(candidate) or not _r5tu_supports_zstd(candidate):
        raise RuntimeError("r5tu binary not found after build")
    return candidate.resolve()


def _r5tu_supports_zstd(r5tu_path: Path) -> bool:
    """Try a tiny build-graph with --zstd to verify feature is enabled."""
    try:
        with tempfile.TemporaryDirectory(prefix="r5tu_zstd_check_") as tdir:
            tdirp = Path(tdir)
            ttl = tdirp / "t.ttl"
            out = tdirp / "t.r5tu"
            ttl.write_text("""
@prefix ex: <http://example.org/> .
ex:s ex:p ex:o .
""".strip())
            cmd = [
                r5tu_path.as_posix(),
                "build-graph",
                "--input",
                ttl.as_posix(),
                "--output",
                out.as_posix(),
                "--format",
                "turtle",
                "--zstd",
            ]
            proc = subprocess.run(cmd, capture_output=True, text=True)
            if proc.returncode != 0:
                combined = (proc.stdout or "") + (proc.stderr or "")
                return "zstd feature not enabled" not in combined
            return out.exists() and out.stat().st_size > 0
    except Exception:
        return False


def iter_ttl_files(root: Path, recursive: bool) -> Iterable[Path]:
    exts = {".ttl", ".turtle"}
    if recursive:
        for p in root.rglob("*"):
            if p.is_file() and p.suffix.lower() in exts:
                yield p
    else:
        for p in root.iterdir():
            if p.is_file() and p.suffix.lower() in exts:
                yield p


def measure_rdflib(ttl_path: Path, out_dir: Path, trials: int = 5) -> Tuple[List[float], List[float], int]:
    try:
        from rdflib import Graph
    except Exception as e:  # pragma: no cover
        raise RuntimeError(
            "rdflib is required. Install with: pip install rdflib"
        ) from e

    load_times: List[float] = []
    save_times: List[float] = []
    out_ttl_size = 0

    for i in range(trials):
        # Load
        g = Graph()
        t0 = time.perf_counter()
        g.parse(ttl_path.as_posix(), format="turtle")
        load_times.append(time.perf_counter() - t0)

        # Serialize
        gc.collect()
        out_ttl = out_dir / f"{ttl_path.stem}.rdflib.t{i}.out.ttl"
        t1 = time.perf_counter()
        g.serialize(destination=out_ttl.as_posix(), format="turtle")
        save_times.append(time.perf_counter() - t1)
        if out_ttl.exists():
            out_ttl_size = out_ttl.stat().st_size

        # Free graph for next iteration
        del g
        gc.collect()

    return load_times, save_times, out_ttl_size


def run_timed(cmd: List[str]) -> Tuple[float, int, str]:
    """Run command, return (seconds, exit_code, stderr+stdout)."""
    t0 = time.perf_counter()
    proc = subprocess.run(cmd, capture_output=True, text=True)
    dt = time.perf_counter() - t0
    out = (proc.stdout or "") + (proc.stderr or "")
    return dt, proc.returncode, out


def measure_r5tu(
    r5tu: Path, ttl_path: Path, out_dir: Path, trials: int = 5
) -> Tuple[List[float], List[float], int, Optional[str]]:
    # Build .r5tu
    build_times: List[float] = []
    stat_times: List[float] = []
    r5_size: int = 0
    last_stat_err: Optional[str] = None

    for i in range(trials):
        out_r5 = out_dir / f"{ttl_path.stem}.trial{i}.r5tu"
        if out_r5.exists():
            try:
                out_r5.unlink()
            except Exception:
                pass
        build_cmd = [
            r5tu.as_posix(),
            "build-graph",
            "--input",
            ttl_path.as_posix(),
            "--output",
            out_r5.as_posix(),
            "--format",
            "turtle",
            "--zstd",
        ]
        build_s, code, build_out = run_timed(build_cmd)
        if code != 0:
            raise RuntimeError(
                f"r5tu build-graph failed (code={code}) for {ttl_path}:\n{build_out}"
            )
        if not out_r5.exists() or out_r5.stat().st_size == 0:
            raise RuntimeError(
                f"r5tu did not produce a valid file: {out_r5} (check that r5tu supports oxigraph CLI)"
            )
        build_times.append(build_s)
        r5_size = out_r5.stat().st_size

        # Stat (load) time
        stat_cmd = [r5tu.as_posix(), "stat", "--file", out_r5.as_posix()]
        stat_s, code, stat_out = run_timed(stat_cmd)
        if code != 0:
            last_stat_err = stat_out
        stat_times.append(stat_s)

    return build_times, stat_times, r5_size, last_stat_err


def main() -> int:
    ap = argparse.ArgumentParser(description="Benchmark Turtle vs rdf5d")
    ap.add_argument("dir", type=Path, help="Directory containing .ttl/.turtle files")
    ap.add_argument(
        "-o",
        "--output",
        type=Path,
        default=Path("benchmark_results.csv"),
        help="CSV output file",
    )
    ap.add_argument(
        "--recursive",
        action="store_true",
        help="Recurse into subdirectories",
    )
    ap.add_argument(
        "--keep-artifacts",
        action="store_true",
        help="Keep generated .r5tu and rdflib .ttl outputs",
    )
    ap.add_argument(
        "--r5tu",
        type=Path,
        default=None,
        help="Path to r5tu binary (if not provided, attempts to find/build)",
    )
    ap.add_argument(
        "--enable-mmap",
        action="store_true",
        help="Build r5tu with mmap feature for mmap-based loading",
    )
    args = ap.parse_args()

    root = args.dir
    if not root.exists() or not root.is_dir():
        print(f"Error: '{root}' is not a directory", file=sys.stderr)
        return 2

    try:
        r5tu_bin = find_or_build_r5tu(args.r5tu, args.enable_mmap)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 2

    ttl_files = list(iter_ttl_files(root, args.recursive))
    if not ttl_files:
        print("No .ttl/.turtle files found", file=sys.stderr)
        return 1

    # Work dir for outputs
    if args.keep_artifacts:
        out_root = Path("bench_artifacts")
        out_root.mkdir(parents=True, exist_ok=True)
        temp_mgr = None
    else:
        temp_mgr = tempfile.TemporaryDirectory(prefix="rdf5d_bench_")
        out_root = Path(temp_mgr.name)

    print(f"Using r5tu: {r5tu_bin}")
    print(f"Artifacts dir: {out_root}")
    results: List[dict] = []

    for i, ttl in enumerate(ttl_files, 1):
        try:
            print(f"[{i}/{len(ttl_files)}] {ttl}")
            ttl_size = ttl.stat().st_size
            file_out_dir = out_root / ttl.stem
            file_out_dir.mkdir(parents=True, exist_ok=True)

            # rdflib (repeat trials)
            rdflib_load_list, rdflib_save_list, rdflib_out_size = measure_rdflib(ttl, file_out_dir, trials=5)

            # rdf5d via r5tu (repeat trials)
            r5_build_list, r5_stat_list, r5_size, stat_err = measure_r5tu(r5tu_bin, ttl, file_out_dir, trials=5)

            # Compute means and stddevs
            rdflib_load_mean = stats.mean(rdflib_load_list)
            rdflib_load_std = stats.stdev(rdflib_load_list) if len(rdflib_load_list) > 1 else 0.0
            rdflib_save_mean = stats.mean(rdflib_save_list)
            rdflib_save_std = stats.stdev(rdflib_save_list) if len(rdflib_save_list) > 1 else 0.0
            r5_build_mean = stats.mean(r5_build_list)
            r5_build_std = stats.stdev(r5_build_list) if len(r5_build_list) > 1 else 0.0
            r5_stat_mean = stats.mean(r5_stat_list)
            r5_stat_std = stats.stdev(r5_stat_list) if len(r5_stat_list) > 1 else 0.0

            rdflib_total_list = [a + b for a, b in zip(rdflib_load_list, rdflib_save_list)]
            r5_total_list = [a + b for a, b in zip(r5_stat_list, r5_build_list)]
            rdflib_total_mean = stats.mean(rdflib_total_list)
            rdflib_total_std = stats.stdev(rdflib_total_list) if len(rdflib_total_list) > 1 else 0.0
            r5_total_mean = stats.mean(r5_total_list)
            r5_total_std = stats.stdev(r5_total_list) if len(r5_total_list) > 1 else 0.0

            save_speedup = (rdflib_save_mean / r5_build_mean) if r5_build_mean > 0 else 0.0
            load_speedup = (rdflib_load_mean / r5_stat_mean) if r5_stat_mean > 0 else 0.0
            total_speedup = (rdflib_total_mean / r5_total_mean) if r5_total_mean > 0 else 0.0
            row = {
                "file": ttl.as_posix(),
                "ttl_size": ttl_size,
                "rdflib_load_mean": rdflib_load_mean,
                "rdflib_load_std": rdflib_load_std,
                "rdflib_save_mean": rdflib_save_mean,
                "rdflib_save_std": rdflib_save_std,
                "rdflib_out_size": rdflib_out_size,
                "r5tu_build_mean": r5_build_mean,
                "r5tu_build_std": r5_build_std,
                "r5tu_load_mean": r5_stat_mean,
                "r5tu_load_std": r5_stat_std,
                "rdflib_total_mean": rdflib_total_mean,
                "rdflib_total_std": rdflib_total_std,
                "r5tu_total_mean": r5_total_mean,
                "r5tu_total_std": r5_total_std,
                "r5tu_size": r5_size,
                "size_ratio_r5tu_over_ttl": (r5_size / ttl_size) if ttl_size else 0.0,
                "size_ratio_r5tu_over_rdflib_out": (r5_size / rdflib_out_size) if rdflib_out_size else 0.0,
                "save_speedup_r5tu_over_rdflib": save_speedup,
                "load_speedup_r5tu_over_rdflib": load_speedup,
                "total_speedup_r5tu_over_rdflib": total_speedup,
            }
            if stat_err:
                row["r5tu_stat_error"] = stat_err.strip().splitlines()[-1][:200]
            results.append(row)
        except Exception as e:
            results.append({
                "file": ttl.as_posix(),
                "error": str(e),
            })

    # Write CSV
    fieldnames = [
        "file",
        "ttl_size",
        "r5tu_size",
        "rdflib_out_size",
        "rdflib_load_mean",
        "rdflib_load_std",
        "rdflib_save_mean",
        "rdflib_save_std",
        "r5tu_build_mean",
        "r5tu_build_std",
        "r5tu_load_mean",
        "r5tu_load_std",
        "rdflib_total_mean",
        "rdflib_total_std",
        "r5tu_total_mean",
        "r5tu_total_std",
        "size_ratio_r5tu_over_ttl",
        "size_ratio_r5tu_over_rdflib_out",
        "save_speedup_r5tu_over_rdflib",
        "load_speedup_r5tu_over_rdflib",
        "total_speedup_r5tu_over_rdflib",
        "r5tu_stat_error",
        "error",
    ]
    # Only include columns present
    present = [f for f in fieldnames if any(f in r for r in results)]
    with args.output.open("w", newline="") as f:
        w = csv.DictWriter(f, fieldnames=present)
        w.writeheader()
        for r in results:
            # Format floats to 6 decimals for readability
            row_out = {}
            for k in present:
                v = r.get(k, "")
                if isinstance(v, float):
                    row_out[k] = f"{v:.6f}"
                else:
                    row_out[k] = v
            w.writerow(row_out)

    print(f"Wrote results to {args.output}")

    # Summary
    ok_rows = [r for r in results if "error" not in r]
    n_ok = len(ok_rows)
    n_err = len(results) - n_ok
    if n_ok:
        total_ttl = sum(r.get("ttl_size", 0) for r in ok_rows)
        total_r5 = sum(r.get("r5tu_size", 0) for r in ok_rows)
        total_rdflib_out = sum(r.get("rdflib_out_size", 0) for r in ok_rows)

        mean_ratio_vs_ttl = sum(r.get("size_ratio_r5tu_over_ttl", 0.0) for r in ok_rows) / n_ok
        mean_ratio_vs_rdflib = (
            sum(r.get("size_ratio_r5tu_over_rdflib_out", 0.0) for r in ok_rows) / n_ok
        ) if total_rdflib_out else 0.0

        mean_rdflib_total = sum(r.get("rdflib_total_mean", 0.0) for r in ok_rows) / n_ok
        mean_r5_total = sum(r.get("r5tu_total_mean", 0.0) for r in ok_rows) / n_ok
        mean_total_speedup = (
            sum(r.get("total_speedup_r5tu_over_rdflib", 0.0) for r in ok_rows) / n_ok
        )

        mean_rdflib_load = sum(r.get("rdflib_load_mean", 0.0) for r in ok_rows) / n_ok
        mean_r5_load = sum(r.get("r5tu_load_mean", 0.0) for r in ok_rows) / n_ok
        mean_load_speedup = (
            sum(r.get("load_speedup_r5tu_over_rdflib", 0.0) for r in ok_rows) / n_ok
        )

        mean_rdflib_save = sum(r.get("rdflib_save_mean", 0.0) for r in ok_rows) / n_ok
        mean_r5_save = sum(r.get("r5tu_build_mean", 0.0) for r in ok_rows) / n_ok
        mean_save_speedup = (
            sum(r.get("save_speedup_r5tu_over_rdflib", 0.0) for r in ok_rows) / n_ok
        )

        pct_saved_vs_ttl = (1 - (total_r5 / total_ttl)) * 100 if total_ttl else 0.0
        pct_saved_vs_rdflib = (
            (1 - (total_r5 / total_rdflib_out)) * 100 if total_rdflib_out else 0.0
        )

        print("\nSummary (rdf5d vs rdflib):")
        print(f"  Files processed: {n_ok} ok, {n_err} errors")
        print(f"  Size total (TTL → R5TU): {total_ttl} → {total_r5} bytes ({pct_saved_vs_ttl:.2f}% saved)")
        if total_rdflib_out:
            print(
                f"  Size total (rdflib TTL → R5TU): {total_rdflib_out} → {total_r5} bytes ({pct_saved_vs_rdflib:.2f}% saved)"
            )
        print(f"  Mean size ratio R5TU/TTL: {mean_ratio_vs_ttl:.3f}")
        if total_rdflib_out:
            print(f"  Mean size ratio R5TU/rdflib TTL: {mean_ratio_vs_rdflib:.3f}")
        print(
            f"  Load+Save total mean (rdflib vs r5tu): {mean_rdflib_total:.3f}s vs {mean_r5_total:.3f}s (speedup {mean_total_speedup:.2f}×)"
        )
        print(
            f"  Load mean (rdflib vs r5tu): {mean_rdflib_load:.3f}s vs {mean_r5_load:.3f}s (speedup {mean_load_speedup:.2f}×)"
        )
        print(
            f"  Save mean (rdflib vs r5tu): {mean_rdflib_save:.3f}s vs {mean_r5_save:.3f}s (speedup {mean_save_speedup:.2f}×)"
        )
    else:
        print("\nSummary: no successful runs to summarize")
    if not args.keep_artifacts and temp_mgr is not None:
        temp_mgr.cleanup()
    return 0


if __name__ == "__main__":
    sys.exit(main())



import unittest
import os
import shutil
from pathlib import Path
from ontoenv import OntoEnv
from rdflib import Graph, URIRef
from rdflib.namespace import RDF, OWL




class TestOntoEnvAPI(unittest.TestCase):
    def setUp(self):
        """Set up a test environment."""
        self.test_dir = Path("test_env_py")
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        self.test_dir.mkdir()

        self.brick_file_path = Path("../brick/Brick.ttl")
        self.brick_name = "https://brickschema.org/schema/1.4-rc1/Brick"
        self.brick_144_url = "https://brickschema.org/schema/1.4.4/Brick.ttl"
        self.brick_144_name = "https://brickschema.org/schema/1.4/Brick"
        self.env = None

        # clean up any existing env in current dir
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def tearDown(self):
        """Tear down the test environment."""
        if self.env:
            self.env.close()
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def test_constructor_default(self):
        """Test default OntoEnv() constructor respects git-style discovery."""
        with self.assertRaises(ValueError):
            OntoEnv()
        self.env = OntoEnv(temporary=True)
        self.assertIn("OntoEnv", repr(self.env))
        
    def test_constructor_path(self):
        """Test OntoEnv(path=...) constructor."""
        self.env = OntoEnv(path=self.test_dir, recreate=True)
        self.assertTrue((self.test_dir / ".ontoenv").is_dir())

    def test_constructor_with_config(self):
        """Test OntoEnv(...flags...) constructor."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["../brick"])
        self.env.update()  # discover ontologies
        ontologies = self.env.get_ontology_names()
        self.assertIn(self.brick_name, ontologies)

    def test_add_local_file(self):
        """Test env.add() with a local file and fetching imports."""
        # requires offline=False to fetch QUDT from web
        self.env = OntoEnv(path=self.test_dir, recreate=True, offline=False)
        name = self.env.add(str(self.brick_file_path))
        self.assertEqual(name, self.brick_name)
        ontologies = self.env.get_ontology_names()
        self.assertIn(self.brick_name, ontologies)
        # check that dependencies were added because fetch_imports is true by default
        self.assertIn("http://qudt.org/2.1/schema/qudt", ontologies)

    def test_add_url(self):
        """Test env.add() with a URL."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, offline=False)
        name = self.env.add(self.brick_144_url)
        self.assertEqual(name, self.brick_144_name)
        ontologies = self.env.get_ontology_names()
        self.assertIn(self.brick_144_name, ontologies)
        # check that dependencies were added because fetch_imports is true by default
        self.assertIn("http://qudt.org/3.1.0/schema/qudt", ontologies)

    def test_add_no_fetch_imports(self):
        """Test env.add() with fetch_imports=False."""
        self.env = OntoEnv(path=self.test_dir, recreate=True)
        # With fetch_imports=False, Brick should be added but its dependencies
        # should not be processed.
        name = self.env.add(str(self.brick_file_path), fetch_imports=False)
        self.assertEqual(name, self.brick_name)
        ontologies = self.env.get_ontology_names()
        self.assertIn(self.brick_name, ontologies)
        # check that dependencies were not added
        self.assertEqual(len(ontologies), 1)

    def test_get_graph(self):
        """Test env.get_graph()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True)
        name = self.env.add(str(self.brick_file_path))
        g = self.env.get_graph(name)
        self.assertIsInstance(g, Graph)
        self.assertGreater(len(g), 0)
        self.assertIn((URIRef(self.brick_name), RDF.type, OWL.Ontology), g)

    def test_get_closure(self):
        """Test env.get_closure()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["brick"])
        name = self.env.add(str(self.brick_file_path))
        g = self.env.get_graph(name)
        closure_g, imported_graphs = self.env.get_closure(name, recursion_depth=0)
        self.assertIsInstance(closure_g, Graph)
        self.assertEqual(len(imported_graphs), 1)

        closure_g, imported_graphs = self.env.get_closure(name)
        self.assertIsInstance(closure_g, Graph)
        self.assertGreater(len(imported_graphs), 1)
        self.assertGreater(len(closure_g), len(g))

    def test_import_dependencies(self):
        """Test env.import_dependencies()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["brick"])
        self.env.add(str(self.brick_file_path))

        g = Graph()
        brick_ontology_uri = URIRef(self.brick_name)
        g.add((brick_ontology_uri, RDF.type, OWL.Ontology))
        # add an import to be removed
        g.add((brick_ontology_uri, OWL.imports, URIRef("http://qudt.org/2.1/schema/qudt")))

        num_triples_before = len(g)
        imported = self.env.import_dependencies(g)
        self.assertGreater(len(imported), 0)
        num_triples_after = len(g)

        self.assertGreater(num_triples_after, num_triples_before)

    def test_import_dependencies_fetch_missing(self):
        """Test env.import_dependencies() with fetch_missing=True."""
        # offline=False is required to fetch from URL
        # empty env
        self.env = OntoEnv(path=self.test_dir, recreate=True, offline=False)
        
        g = Graph()
        # Add an import to a known ontology URL that is not in the environment
        g.add(
            (
                URIRef("http://example.org/my-ontology"),
                OWL.imports,
                URIRef(self.brick_144_url),
            )
        )

        num_triples_before = len(g)
        # With fetch_missing=True, this should download Brick and its dependencies
        imported = self.env.import_dependencies(g, fetch_missing=True)
        self.assertGreater(len(imported), 0)
        self.assertIn(self.brick_144_name, imported)
        num_triples_after = len(g)

        self.assertGreater(num_triples_after, num_triples_before)

        # check that the fetched ontologies are now in the environment
        ontologies = self.env.get_ontology_names()
        self.assertIn(self.brick_144_name, ontologies)

    def test_list_closure(self):
        """Test env.list_closure()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["brick"])
        name = self.env.add(str(self.brick_file_path))
        closure_list = self.env.list_closure(name)
        self.assertIn(name, closure_list)
        # check for some known imports
        self.assertIn("http://qudt.org/2.1/schema/qudt", closure_list)
        self.assertIn("http://qudt.org/2.1/vocab/quantitykind", closure_list)

    def test_get_importers(self):
        """Test env.get_importers()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["brick"])
        self.env.add(str(self.brick_file_path))

        dependents = self.env.get_importers("http://qudt.org/2.1/vocab/quantitykind")
        self.assertIn(self.brick_name, dependents)

    def test_to_rdflib_dataset(self):
        """Test env.to_rdflib_dataset()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["brick"])
        self.env.add(str(self.brick_file_path))
        self.env.update()  # need to run update to find all dependencies
        self.env.flush()

        ds = self.env.to_rdflib_dataset()
        # count graphs
        num_graphs = len(list(ds.graphs()))
        # there should be many graphs: brick + all imports
        self.assertGreater(num_graphs, 5)

    def test_import_graph(self):
        """Test env.import_graph()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, offline=False)
        name = self.env.add(self.brick_144_url)
        self.assertEqual(name, self.brick_144_name)

        g = Graph()
        self.assertEqual(len(g), 0)
        self.env.import_graph(g, name)
        self.assertGreater(len(g), 0)

    def test_store_path(self):
        """Test env.store_path()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True)
        path = self.env.store_path()
        self.assertIsNotNone(path)
        self.assertTrue(Path(path).is_dir())
        self.assertIn(".ontoenv", path)

        # for in-memory, it should be None
        mem_env = OntoEnv(temporary=True)
        self.assertIsNone(mem_env.store_path())
        mem_env.close()

    def test_persistence(self):
        """Test that the environment is persisted to disk."""
        env = OntoEnv(path=self.test_dir, recreate=True)
        name = env.add(str(self.brick_file_path))
        self.assertIn(name, env.get_ontology_names())
        env.flush()  # ensure everything is written to disk
        env.close()

        # load it again from the same path
        self.env = OntoEnv(path=self.test_dir)
        self.assertIn(name, self.env.get_ontology_names())
        g = self.env.get_graph(name)
        self.assertGreater(len(g), 0)

    def test_close(self):
        """Test that the environment can be closed and methods fail."""
        self.env = OntoEnv(path=self.test_dir, recreate=True)
        name = self.env.add(str(self.brick_file_path))
        self.assertIn(name, self.env.get_ontology_names())
        self.env.close()

        # check that methods raise a ValueError
        with self.assertRaises(ValueError):
            self.env.get_ontology_names()
        with self.assertRaises(ValueError):
            self.env.get_graph(name)
        with self.assertRaises(ValueError):
            self.env.add(str(self.brick_file_path))

        # check __repr__
        self.assertIn("closed", repr(self.env))

        # store path should be None
        self.assertIsNone(self.env.store_path())

        # closing again should be fine
        self.env.close()

        # check that we can still create a new env from the same directory,
        # which should load the persisted state.
        env2 = OntoEnv(path=self.test_dir)
        self.assertIn(name, env2.get_ontology_names())
        env2.close()

    def test_get_dependencies_graph(self):
        """Test env.get_dependencies_graph()."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, offline=False)
        self.env.add(str(self.brick_file_path))

        g = Graph()
        brick_ontology_uri = URIRef(self.brick_name)
        g.add((brick_ontology_uri, RDF.type, OWL.Ontology))
        # add an import to be resolved
        g.add((brick_ontology_uri, OWL.imports, URIRef("http://qudt.org/2.1/vocab/quantitykind")))

        num_triples_before = len(g)
        deps_g, imported = self.env.get_dependencies_graph(g)
        num_triples_after = len(g)

        # original graph should not be modified
        self.assertEqual(num_triples_before, num_triples_after)

        # new graph should have content
        self.assertGreater(len(deps_g), 0)
        self.assertGreater(len(imported), 0)
        self.assertIn("http://qudt.org/2.1/vocab/quantitykind", imported)
        self.assertIn("http://qudt.org/2.1/vocab/dimensionvector", imported)

        # test with destination graph
        dest_g = Graph()
        self.assertEqual(len(dest_g), 0)
        deps_g2, imported2 = self.env.get_dependencies_graph(g, destination_graph=dest_g)

        # check that the returned graph is the same object as the destination graph
        self.assertIs(deps_g2, dest_g)
        self.assertGreater(len(dest_g), 0)
        self.assertEqual(len(deps_g), len(dest_g))
        self.assertEqual(sorted(imported), sorted(imported2))

    def test_update_all_flag(self):
        """Test env.update(all=True) forces reloading of all ontologies."""
        self.env = OntoEnv(path=self.test_dir, recreate=True, search_directories=["../brick"])
        # Initial discovery of ontologies
        self.env.update()
        self.assertIn(self.brick_name, self.env.get_ontology_names())

        ont1 = self.env.get_ontology(self.brick_name)
        ts1 = ont1.last_updated
        self.assertIsNotNone(ts1)

        # Force update of all ontologies
        self.env.update(all=True)

        ont2 = self.env.get_ontology(self.brick_name)
        ts2 = ont2.last_updated
        self.assertIsNotNone(ts2)
        self.assertNotEqual(ts1, ts2)



if __name__ == "__main__":
    unittest.main()



"""Python package shim for the ontoenv extension."""

# Try both common names to tolerate different build configurations
try:  # prefer the extension named 'ontoenv'
    from .ontoenv import *  # type: ignore[attr-defined]
    from . import ontoenv as _ext  # type: ignore[attr-defined]
except Exception:  # fallback to '_ontoenv'
    from ._ontoenv import *  # type: ignore[attr-defined]
    from . import _ontoenv as _ext  # type: ignore[attr-defined]

__doc__ = getattr(_ext, "__doc__", None)
if hasattr(_ext, "__all__"):
    __all__ = _ext.__all__  # type: ignore[attr-defined]



from __future__ import annotations

import sys

from .ontoenv import run_cli as _run_cli


def main(argv: list[str] | None = None) -> int:
    code = _run_cli(argv if argv is not None else list(sys.argv))
    if code != 0:
        raise SystemExit(code)
    return 0



import unittest
import shutil
import multiprocessing
from pathlib import Path
from rdflib import Graph, URIRef
from rdflib.namespace import RDF, OWL
from ontoenv import OntoEnv


def _ro_open_worker(path_str, graph_uri, result_queue):
    try:
        from pathlib import Path
        from ontoenv import OntoEnv
        from rdflib import URIRef
        from rdflib.namespace import RDF, OWL
        import time

        # Open the same store in read-only mode and load a specific graph + metadata
        env = OntoEnv(path=Path(path_str), read_only=True)
        g = env.get_graph(graph_uri)
        ok_graph = (URIRef(graph_uri), RDF.type, OWL.Ontology) in g and len(g) > 0
        ont = env.get_ontology(graph_uri)
        ok_meta = (ont.id == graph_uri)

        # Brief sleep to increase overlap
        time.sleep(0.2)
        env.close()
        result_queue.put(("ok", graph_uri) if (ok_graph and ok_meta) else ("missing", graph_uri))
    except Exception as e:
        result_queue.put(("error", graph_uri, str(e)))


def _rw_open_worker(path_str, graph_uri, result_queue):
    try:
        from pathlib import Path
        from ontoenv import OntoEnv
        from rdflib import URIRef
        from rdflib.namespace import RDF, OWL
        import time

        # Open the same store in read-write mode and load a specific graph + metadata
        env = OntoEnv(path=Path(path_str))
        g = env.get_graph(graph_uri)
        ok_graph = (URIRef(graph_uri), RDF.type, OWL.Ontology) in g and len(g) > 0
        ont = env.get_ontology(graph_uri)
        ok_meta = (ont.id == graph_uri)

        time.sleep(1.0)
        env.close()
        result_queue.put(("ok", graph_uri) if (ok_graph and ok_meta) else ("missing", graph_uri))
    except Exception as e:
        result_queue.put(("error", graph_uri, str(e)))


def _writer_hold_worker(path_str, hold_secs, graph_uri, result_queue):
    try:
        import time
        from pathlib import Path
        from ontoenv import OntoEnv
        from rdflib import URIRef
        from rdflib.namespace import RDF, OWL

        env = OntoEnv(path=Path(path_str))
        # Touch a known graph to ensure the store is usable
        g = env.get_graph(graph_uri)
        ok_graph = (URIRef(graph_uri), RDF.type, OWL.Ontology) in g and len(g) > 0

        # Hold the exclusive writer lock for a bit
        time.sleep(hold_secs)
        env.close()
        result_queue.put(("released", ok_graph))
    except Exception as e:
        result_queue.put(("error", str(e)))


def _ro_open_get_graph_worker(path_str, graph_uri, result_queue):
    try:
        from pathlib import Path
        from ontoenv import OntoEnv
        from rdflib import URIRef
        from rdflib.namespace import RDF, OWL

        env = OntoEnv(path=Path(path_str), read_only=True)
        g = env.get_graph(graph_uri)
        ok = (URIRef(graph_uri), RDF.type, OWL.Ontology) in g and len(g) > 0
        env.close()
        result_queue.put(("ok", ok))
    except Exception as e:
        result_queue.put(("error", str(e)))

class TestOntoEnvReadOnlyConcurrency(unittest.TestCase):
    def setUp(self):
        self.test_dir = Path("test_env_ro")
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        self.test_dir.mkdir()
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def tearDown(self):
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def test_concurrent_read_only_open_same_store(self):
        # Pre-create a persistent store with two different ontologies
        a_path = self.test_dir / "A.ttl"
        b_path = self.test_dir / "B.ttl"
        a_uri = "http://example.org/ont/A"
        b_uri = "http://example.org/ont/B"
        a_path.write_text(
            "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
            "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
            f"<{a_uri}> a owl:Ontology .\n"
            f"<{a_uri}#Class1> a owl:Class .\n",
            encoding="utf-8",
        )
        b_path.write_text(
            "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
            "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
            f"<{b_uri}> a owl:Ontology .\n"
            f"<{b_uri}#Class2> a owl:Class .\n",
            encoding="utf-8",
        )

        # Create the store and add ontologies (single writer)
        env = OntoEnv(path=self.test_dir, recreate=True)
        name_a = env.add(str(a_path), fetch_imports=False)
        name_b = env.add(str(b_path), fetch_imports=False)
        self.assertEqual(name_a, a_uri)
        self.assertEqual(name_b, b_uri)
        env.flush()
        env.close()

        # Now, concurrently open the same store as read-only from two processes
        ctx = multiprocessing.get_context("spawn")
        q = ctx.Queue()
        store_path = str(self.test_dir.resolve())
        p1 = ctx.Process(target=_ro_open_worker, args=(store_path, name_a, q))
        p2 = ctx.Process(target=_ro_open_worker, args=(store_path, name_b, q))

        p1.start()
        p2.start()

        r1 = q.get(timeout=30)
        r2 = q.get(timeout=30)

        p1.join(timeout=30)
        p2.join(timeout=30)

        self.assertFalse(p1.is_alive())
        self.assertFalse(p2.is_alive())
        self.assertEqual(p1.exitcode, 0)
        self.assertEqual(p2.exitcode, 0)

        results = {r1, r2}
        self.assertIn(("ok", name_a), results)
        self.assertIn(("ok", name_b), results)


class TestOntoEnvRWConcurrency(unittest.TestCase):
    def setUp(self):
        self.test_dir = Path("test_env_py")
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        self.test_dir.mkdir()
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def tearDown(self):
        if self.test_dir.exists():
            shutil.rmtree(self.test_dir)
        if Path(".ontoenv").exists():
            shutil.rmtree(".ontoenv")

    def test_concurrent_open_same_store(self):
        """Attempt to open the same ontoenv store in two different processes, each loading a different graph."""
        # Pre-create a persistent store with two different ontologies
        a_path = self.test_dir / "A.ttl"
        b_path = self.test_dir / "B.ttl"
        a_uri = "http://example.org/ont/A"
        b_uri = "http://example.org/ont/B"
        a_path.write_text(
            "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
            "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
            f"<{a_uri}> a owl:Ontology .\n"
            f"<{a_uri}#Class1> a owl:Class .\n",
            encoding="utf-8",
        )
        b_path.write_text(
            "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
            "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
            f"<{b_uri}> a owl:Ontology .\n"
            f"<{b_uri}#Class2> a owl:Class .\n",
            encoding="utf-8",
        )

        env = OntoEnv(path=self.test_dir, recreate=True)
        name_a = env.add(str(a_path), fetch_imports=False)
        name_b = env.add(str(b_path), fetch_imports=False)
        self.assertEqual(name_a, a_uri)
        self.assertEqual(name_b, b_uri)
        env.flush()
        env.close()

        ctx = multiprocessing.get_context("spawn")
        q = ctx.Queue()
        p1 = ctx.Process(target=_rw_open_worker, args=(str(self.test_dir), name_a, q))
        p2 = ctx.Process(target=_rw_open_worker, args=(str(self.test_dir), name_b, q))

        p1.start()
        p2.start()

        r1 = q.get(timeout=30)
        r2 = q.get(timeout=30)

        p1.join(timeout=30)
        p2.join(timeout=30)

        # Verify both processes finished; one should succeed and one should report a lock error
        self.assertFalse(p1.is_alive())
        self.assertFalse(p2.is_alive())
        self.assertEqual(p1.exitcode, 0)
        self.assertEqual(p2.exitcode, 0)

        results = [r1, r2]
        ok_results = [r for r in results if r[0] == "ok"]
        error_results = [r for r in results if r[0] == "error"]

        # At least one process should fail to acquire the exclusive lock
        self.assertGreaterEqual(len(ok_results), 1)
        self.assertGreaterEqual(len(error_results), 1)

        # The successful open(s) should be for one of the graphs we added
        for tag, uri in ok_results:
            self.assertIn(uri, {name_a, name_b})

        # The error should mention failure to acquire exclusive lock for write
        err_msg = error_results[0][2]
        self.assertTrue(
            "Failed to open OntoEnv store for write" in err_msg or "exclusive lock" in err_msg,
            msg=f"Unexpected error message: {err_msg}",
        )

    def test_reader_waits_for_writer_then_reads(self):
        """A read-only open should wait while a writer holds the exclusive lock, then succeed."""
        a_path = self.test_dir / "A.ttl"
        a_uri = "http://example.org/ont/A"
        a_path.write_text(
            "@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .\n"
            "@prefix owl: <http://www.w3.org/2002/07/owl#> .\n"
            f"<{a_uri}> a owl:Ontology .\n",
            encoding="utf-8",
        )
        env = OntoEnv(path=self.test_dir, recreate=True)
        env.add(str(a_path), fetch_imports=False)
        env.flush()
        env.close()

        ctx = multiprocessing.get_context("spawn")
        q = ctx.Queue()
        hold_secs = 1.0
        writer = ctx.Process(target=_writer_hold_worker, args=(str(self.test_dir), hold_secs, a_uri, q))
        reader = ctx.Process(target=_ro_open_get_graph_worker, args=(str(self.test_dir), a_uri, q))

        writer.start()
        import time
        time.sleep(0.15)  # ensure writer started and holds the lock
        t0 = time.time()
        reader.start()

        # First result should be from reader after writer releases
        r1 = q.get(timeout=30)
        r2 = q.get(timeout=30)

        writer.join(timeout=30)
        reader.join(timeout=30)

        self.assertFalse(writer.is_alive())
        self.assertFalse(reader.is_alive())
        self.assertEqual(writer.exitcode, 0)
        self.assertEqual(reader.exitcode, 0)

        elapsed = time.time() - t0
        # Reader should have waited roughly the hold duration (minus start skew)
        self.assertGreaterEqual(elapsed, 0.7)

        results = {r1[0], r2[0]}
        self.assertIn("released", results)
        self.assertIn("ok", results)


if __name__ == "__main__":
    unittest.main()
